/*
 * Copyright (c)  2007 - 2020 by inovat
 * innovative systeme - verkehr - tunnel - technik,
 * Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator
 *
 * de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Diese Datei ist Teil von de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator.
 *
 * de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 * de.bsvrz.sys.funclib.kbgen.ConfigAreaCreator wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 *
 * Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 * Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */


package de.bsvrz.sys.funclib.kbgen;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.sys.funclib.debug.Debug;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Erzeugt aus einer Liste von Objekten vom Typ {@link ProxyConfigurationObject} die Objektdefinitionen für einen Konfigurationsbereich im
 * Datenkatalog.
 * <p>
 * Dazu enthält ein {@link ProxyConfigurationObject} alle Informationen zu dem zu erzeugenden Konfigurationsobjekt. Dies sind neben Typ, Name und Pid
 * insbesondere die Informationen zu den konfigurierenden Attributgruppen. Damit die Klasse {@link ConfigAreaCreator} die Informationen zu den
 * konfigurierenden Attributgruppen genenerisch ermitteln kann, halt ein {@link ProxyConfigurationObject} zu jeder konfigurierenden Attributgruppe die
 * Daten in Form einer {@link de.bsvrz.dav.daf.main.Data} Struktur, über die generisch iteriert werden kann. Das Problem besteht nun darin, dass
 * Objektreferenzen in einer solchen {@link de.bsvrz.dav.daf.main.Data} Struktur bei der Erstellung des {@link ProxyConfigurationObject} nicht
 * eingetragen werden können, da dass referenzierte Objekt i. d. R. noch gar nicht in der Konfiguration verfügbar ist. Der Versuch eine solche
 * Referenz anzulegen führt deshalb zu einem NullPointerFehler. Aus diesem Grund wird zu einem {@link de.bsvrz.dav.daf.main.Data} Objekt zusätzlich
 * eine Liste mit den Pid der zu referenzierenden Objekte mitgeführt, wobei die Reihenfolge der Einträge in der Liste der Reihenfolge der
 * Objektreferenzen im {@link de.bsvrz.dav.daf.main.Data} Objekt entspricht. Bei der Auswertung der Informationen durch die Klasse {@link
 * ConfigAreaCreator} ermittelt diese dann den Wert einer Referenz nicht aus dem {@link de.bsvrz.dav.daf.main.Data} Objekt, sondern aus dem
 * entsprechenden Eintrag der Liste.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 */
public class ConfigAreaCreator {

    /** DebugLogger für Debug-Ausgaben. */
    private static final Debug DEBUG = Debug.getLogger();

    /** Ausgabestream für den zu erstellenden Konfigurationsbereich. */
    private final BufferedOutputStream _bos;

    /** Die Konfigurationsdaten, die in den Konfigurationsbereich umgesetzt werden sollen. */
    private final List<ProxyConfigurationObject> _configAreaData;

    /** File, in das der erzeugte Konfiguraitonsbereich geschrieben wird. */
    private final File _configAreaFile;

    /** Infotext für diesen Konfigurationsbereich. */
    private final String _configAreaInfo;

    /** Name des Konfigurationsbereichs. */
    private final String _configAreaName;

    /** PID für den Konfigurationsbereich. */
    private final String _configAreaPid;

    /** Konfigurationsverantwortlicher (KV). */
    private final String _configResponsibilityName;

    /**
     * Erzeugt ein Objekt vom Typ ConfigAreaCreator.
     *
     * @param configAreaFile           Datei, in das der erzeugte Konfigurationsbereich geschrieben wird.
     * @param configAreaPid            PID für den Konfigurationsbereich.
     * @param configAreaName           Name des Konfigurationsbereichs.
     * @param configAreaInfo           Infotext für diesen Konfigurationsbereich.
     * @param configResponsibilityName Konfigurationsverantwortlicher.
     * @param configAreaData           Die Konfigurationsdaten, die in den Konfigurationsbereich umgesetzt werden sollen.
     */
    public ConfigAreaCreator(File configAreaFile, String configAreaPid, String configAreaName, String configAreaInfo, String configResponsibilityName,
                             List<ProxyConfigurationObject> configAreaData) {
        _configAreaFile = configAreaFile;

        // Konfigurationsdaten umkopieren, da Liste manipuliert wird.
        _configAreaData = new ArrayList<>();
        _configAreaData.addAll(configAreaData);
        _configAreaPid = configAreaPid;
        _configAreaName = configAreaName;
        _configAreaInfo = configAreaInfo;
        _configResponsibilityName = configResponsibilityName;

        try {
            _bos = new BufferedOutputStream(new FileOutputStream(_configAreaFile));
        } catch (IOException e) {
            DEBUG.error("Fehler beim anlegen der Datei: " + _configAreaFile.getAbsoluteFile());
            e.printStackTrace();

            throw new RuntimeException(e);
        }
    }

    /**
     * Erzeugt abhängig vom aktuellen Einrückungslevel eintsprechende Tabulatoreinrückungen.
     *
     * @param level Aktuelle Einrücktiefe.
     *
     * @return Ein StringBuffer mit Tabstopps entspechend level.
     */
    private StringBuffer indent(int level) {
        StringBuffer sb = new StringBuffer(level);

        for (int i = 0; i < level; ++i) {
            sb.append("\t");
        }

        return sb;
    }

    /**
     * Einen Datensatz ausgeben. Grundsätzlich wird unterschieden zwischen "einfachen" Daten (Eigenschaft <code>isPlain</code>) und im Gegensatz dazu
     * Feldern und Listen. Zur Ausgabe der letzteren beiden ruft sich diese Methode rekursiv wieder auf.
     *
     * @param data                          Daten der Konfigurierenden Attributgruppe.
     * @param referenceValues               Liste mit den Referenzen der Attributgruppe.
     * @param pidObjectTypeDefaultParameter PID des Objekttyps, der beim Anlegen des Defaultparameters verwendet werden soll oder null, wenn kein
     *                                      Defaultparameter sondern eine normale ATG angelegt werden soll.
     * @param level                         Aktuelle Einrücktiefe.
     * @param sb                            Ausgabebuffer.
     */
    @SuppressWarnings({"checkstyle:ExecutableStatementCount", "checkstyle:NestedIfDepth"})
    private void printData(Data data, List<String> referenceValues, String pidObjectTypeDefaultParameter, int level, StringBuffer sb) {
        if (data.isPlain()) {

            /* Name dieses Datums */
            String name = data.getName();
            AttributeType attributeType = data.getAttributeType();
            String value;

            if (attributeType instanceof ReferenceAttributeType) {
                value = referenceValues.get(0);
                referenceValues.remove(0);
            } else {

                /* Der konkrete Wert dieses Datums */
                value = data.asTextValue().getValueText();
            }

            // <datum name="name" wert="value"/>
            sb.append(indent(level));
            sb.append("<datum name=\"");
            sb.append(name);
            sb.append("\"" + " wert=\"");
            sb.append(value);
            sb.append("\"/>\n");
        } else if (data.isArray()) {

            /* Iterator zum Durchlaufen des Feldes */
            Iterator<Data> i = data.iterator();

            /* Name dieses Datums */
            String name = data.getName();

            // <datenfeld name="name">
            sb.append(indent(level)).append("<datenfeld name=\"").append(name).append("\">\n");
            level++;

            while (i.hasNext()) {
                printData(i.next(), referenceValues, pidObjectTypeDefaultParameter, level, sb);
            }

            level--;

            // </datenfeld>
            sb.append(indent(level)).append("</datenfeld>\n");
        } else if (data.isList()) {

            /* Iterator zum Durchlaufen der Liste */
            Iterator<Data> i = data.iterator();

            /* Name dieses Datums */
            String name = data.getName();

            if (level == 3) {
                if (pidObjectTypeDefaultParameter == null) {

                    // Öffnendes Tag für Datensatz schreiben
                    sb.append(indent(level)).append("<datensatz attributgruppe=\"").append(name).append("\">\n");
                } else {

                    // Öffnendes Tag für DefaultParameter schreiben
                    // <defaultParameter attributgruppe="atg.protokollEinstellungenStandard" typ="typ.anschlussPunkt">
                    sb.append(indent(level)).append("<defaultParameter attributgruppe=\"").append(name).append("\"").append(" typ=\"")
                        .append(pidObjectTypeDefaultParameter).append("\">\n");
                }
            } else {

                // <datenliste name="name">
                sb.append(indent(level)).append("<datenliste name=\"").append(name).append("\">\n");
            }

            level++;

            while (i.hasNext()) {
                printData(i.next(), referenceValues, pidObjectTypeDefaultParameter, level, sb);
            }

            level--;

            if (level == 3) {
                if (pidObjectTypeDefaultParameter == null) {

                    // Schließendes Tag für Datensatz schreiben
                    sb.append(indent(level)).append("</datensatz>\n");
                } else {

                    // Schließendes Tag für DefaultParameter schreiben
                    sb.append(indent(level)).append("</defaultParameter>\n");
                }
            } else {

                // </datenliste>
                sb.append(indent(level)).append("</datenliste>\n");
            }
        }
    }

    /**
     * Gibt eine Menge eines Konfigurationsobjekts aus.
     *
     * @param proxySetDescription Menge, die ausgegeben werden soll.
     * @param level               Aktuelle Einrücktiefe.
     * @param sb                  Ausgabebuffer.
     */
    private void printSet(ProxySetDescription proxySetDescription, int level, StringBuffer sb) {
        String name = proxySetDescription.getName();

        // <objektMenge name="name">
        sb.append(indent(level)).append("<objektMenge name=\"").append(name).append("\">\n ");
        level++;

        for (String pid : proxySetDescription.getElements()) {

            // <element pid="name"/>
            sb.append(indent(level)).append("<element pid=\"").append(pid).append("\"/>\n");
        }

        level--;

        // </objektMenge>
        sb.append(indent(level)).append("</objektMenge>\n");
    }

    /** Konvertiert die Daten in eine Konfigurationsdatei. */
    @SuppressWarnings("checkstyle:ExecutableStatementCount")
    public void run() {
        StringBuffer sb = new StringBuffer();

        // Konfigurationsbereichskopf anlegen
        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<!DOCTYPE konfigurationsBereich PUBLIC \"-//K2S//DTD Dokument//DE\" \"K2S.dtd\">\n");
        sb.append("<konfigurationsBereich pid=\"");
        sb.append(_configAreaPid);
        sb.append("\" name=\"");
        sb.append(_configAreaName);
        sb.append("\" verantwortlich=\"");
        sb.append(_configResponsibilityName);
        sb.append("\">\n" + "\t<info>\n" + "\t\t<kurzinfo>");
        sb.append(_configAreaInfo);
        sb.append("</kurzinfo>\n" + "\t</info>\n" + "\t<objekte>\n");
        writeConfigData(sb);

        int level = 2;

        try {
            for (ProxyConfigurationObject pco : _configAreaData) {
                sb = new StringBuffer();

                // Öffnendes Tag für Konfigurationsobjekt schreiben
                sb.append(indent(level));
                sb.append("<konfigurationsObjekt typ=\"");

                // System.out.println("pco: " +  pco.toString());
                sb.append(pco.getType().getPid());
                sb.append("\"" + " pid=\"");
                sb.append(pco.getPid());
                sb.append("\"" + " name=\"");
                sb.append(pco.getName());
                sb.append("\">" + "\n");
                level++;

                // Konfigurierende Atg des Konfigurationsobjekts schreiben
                for (ProxyDataDescription proxyDataDescription : pco.getProxyConfigurationDataList()) {
                    Data data = proxyDataDescription.getData();

                    List<String> referenceValues = new ArrayList<>(proxyDataDescription.getReferenceValues());

                    // Datensatz rekursiv ausgeben
                    printData(data, referenceValues, proxyDataDescription.getPidTypeDefaultParameter(), level, sb);
                }

                // Mengen des Konfigurationsobjekts schreiben
                for (ProxySetDescription proxySetDescription : pco.getProxySetList()) {
                    printSet(proxySetDescription, level, sb);
                }

                level--;

                // Schließendes Tag für Konfigurationsobjekt schreiben
                sb.append(indent(level)).append("</konfigurationsObjekt>\n");
                writeConfigData(sb);
            }
        } catch (Exception e) {
            DEBUG.error("Fehler beim Erzeugen der Konfigurationsdatei: ", e);
            e.printStackTrace();

            throw new RuntimeException(e);
        }

        sb = new StringBuffer();
        sb.append("\t</objekte>\n" + "</konfigurationsBereich>");
        writeConfigData(sb);

        try {
            _bos.close();
        } catch (IOException e) {
            DEBUG.error("Fehler beim Schließen der Datei.");
            e.printStackTrace();

            throw new RuntimeException(e);
        }
    }

    /**
     * Gibt die String-Repräsentation dieses Objekts aus.
     * <p>
     * Die Formatierung der Ausgabe kann sich ändern und wird nicht garantiert.
     *
     * @return Aktueller Zustand der Klassenvariablen dieses Objekts.
     */
    public String toString() {
        return "ConfigAreaCreator{" + "_configAreaData=" + _configAreaData + ", _configAreaFile=" + _configAreaFile + ", _configAreaPid='" +
               _configAreaPid + '\'' + ", _configAreaName='" + _configAreaName + '\'' + ", _configAreaInfo='" + _configAreaInfo + '\'' +
               ", _configResponsibilityName='" + _configResponsibilityName + '\'' + ", _bos=" + _bos + '}';
    }

    /**
     * Schreibt einen Stringbuffer in Datei.
     *
     * @param sb Zu schreibender StringBuffer.
     */
    private void writeConfigData(StringBuffer sb) {
        try {
            _bos.write(sb.toString().getBytes());
        } catch (IOException e) {
            DEBUG.error("Fehler beim Schreiben auf Datei: " + _configAreaFile.getAbsoluteFile());
            e.printStackTrace();

            throw new RuntimeException(e);
        }
    }
}

